import os
import shutil
import sys

from dafne_dl import DynamicDLModel


def generate_convert(model_id,
                     default_weights_path,
                     model_name_prefix,
                     model_create_function,
                     model_apply_function,
                     model_learn_function,
                     dimensionality=2,
                     model_type=DynamicDLModel,
                     metadata=None):
    """
    Function that either generates a new model using the default weights, or updates an existing model, based on argv[1]

    Parameters:
        model_id: the id of the model to be generated/updated
        default_weights_path: the path to the default weights
        model_name_prefix: the prefix of the model name (e.g. 'Leg' or 'Thigh'). Used for the filename
        model_create_function: the function that creates the model
        model_apply_function: the function that applies the model
        model_learn_function: the function that performs the incremental learning of the model
        dimensionality: the dimensionality of the data (2 or 3)
        model_type: the type of the model (e.g. DynamicDLModel or DynamicTorchModel)
        metadata: additional metadata to be stored in the model

    Returns:
        None
    """
    if len(sys.argv) > 1:
        # convert an existing model
        print("Converting model", sys.argv[1])
        old_model_path = sys.argv[1]
        filename = old_model_path
        old_model = model_type.Load(open(old_model_path, 'rb'))
        shutil.move(old_model_path, old_model_path + '.bak')
        weights = old_model.get_weights()
        timestamp = old_model.timestamp_id
        model_id = old_model.model_id
    else:
        model_id = model_id
        timestamp = 1610001000
        model = model_create_function()
        try:
            model.load_weights(default_weights_path)
            weights = model.get_weights()
        except AttributeError:
            # model is pytorch, not keras
            import torch
            from dafne_dl.misc import torch_apply_fn_to_state_1
            model.load_state_dict(torch.load(default_weights_path, weights_only=True, map_location=torch.device('cpu')))
            weights = torch_apply_fn_to_state_1(model.state_dict(), lambda x: x.clone())

        filename = f'models/{model_name_prefix}_{timestamp}.model'

    modelObject = model_type(model_id,
                                 model_create_function,
                                 model_apply_function,
                                 incremental_learn_function=model_learn_function,
                                 weights=weights,
                                 timestamp_id=timestamp,
                                 data_dimensionality=dimensionality,
                                 metadata=metadata
                                 )

    with open(filename, 'wb') as f:
        modelObject.dump(f)

    try:
        os.remove(filename + '.sha256')
    except FileNotFoundError:
        pass

    print('Saved', filename)


import ast
import sys

def extract_function_source_basic(source_code, function_name):
    fn_found = False
    fn_source = ''
    for line in source_code.split('\n'):
        if fn_found:
            if not (line.startswith(' ') or line.startswith('\t')) and line.strip() != '':
                # end of function definition
                break
            fn_source += line + '\n'
        if 'def ' + function_name in line:
            fn_found = True
            fn_source = line + '\n'
    return fn_source


def extract_function_source(source_code, function_name):
    """
    Extracts the source code of a specific function from a string containing Python source code.
    Supports nested functions and functions that end with a nested function definition or 'generate_convert' statement.

    Args:
        source_code (str): The Python source code string to extract the function from.
        function_name (str): The name of the function to extract.

    Returns:
        str: The source code string of the requested function, or an empty string if the function was not found.
    """
    class FunctionVisitor(ast.NodeVisitor):
        def __init__(self, function_name):
            self.function_name = function_name
            self.found_function = False
            self.function_source = ""

        def visit_FunctionDef(self, node):
            if node.name == self.function_name:
                self.found_function = True
                self.function_source = ast.unparse(node).strip()
            else:
                self.generic_visit(node)

        def visit_ClassDef(self, node):
            return  # Skip nested class definitions

        def visit_Lambda(self, node):
            return  # Skip lambda functions

    tree = ast.parse(source_code)
    function_visitor = FunctionVisitor(function_name)
    function_visitor.visit(tree)
    if function_visitor.found_function:
        return function_visitor.function_source
    else:
        return ""



import os
import time

from dafne_dl import DynamicDLModel

from PyQt5 import QtWidgets, QtGui, QtCore
from PyQt5.QtCore import pyqtSlot, pyqtSignal, QObject, QVariant

import matplotlib.pyplot as plt
import numpy as np
from PyQt5.QtWidgets import QMessageBox, QFileDialog, QWidget
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.colors import ListedColormap
from matplotlib.ticker import FuncFormatter
from tensorflow.keras.callbacks import Callback

from .ModelTrainer_Ui import Ui_ModelTrainerUI
from ..bin.create_model import load_data, get_model_info, create_model_source, get_model_functions, train_model, \
    prepare_data, set_data_path, set_force_preprocess
from ..utils.ThreadHelpers import separate_thread_decorator

PATIENCE = 10
MIN_EPOCHS = 20

class PredictionUICallback(Callback, QObject):
    fit_signal = pyqtSignal(float, float, np.ndarray, np.ndarray)

    def __init__(self, test_image=None):
        Callback.__init__(self)
        QObject.__init__(self)
        self.min_val_loss = np.inf
        self.n_val_loss_increases = 0
        self.test_image = test_image
        self.do_stop = False
        self.best_weights = None
        self.auto_stop_training = False

    @pyqtSlot(np.ndarray)
    def set_test_image(self, image):
        self.test_image = image

    @pyqtSlot(bool)
    def set_auto_stop_training(self, auto_stop):
        self.auto_stop_training = auto_stop

    @pyqtSlot()
    def stop(self):
        self.do_stop = True
        print('Stopping training...')

    def on_epoch_end(self, epoch, logs=None):
        if self.do_stop:
            self.model.stop_training = True
            return

        if 'val_loss' in logs:
            val_loss = logs['val_loss']
        else:
            val_loss = np.inf

        loss = logs['loss']

        if self.auto_stop_training:
            if epoch >= MIN_EPOCHS and val_loss < self.min_val_loss:
                self.min_val_loss = val_loss
                self.n_val_loss_increases = 0
                self.best_weights = self.model.get_weights()
            elif val_loss > self.min_val_loss:
                self.n_val_loss_increases += 1

            if self.n_val_loss_increases >= PATIENCE:
                self.model.stop_training = True
        else:
            self.best_weights = None
            self.min_val_loss = np.inf

        if self.test_image is None:
            self.fit_signal.emit(loss, val_loss, np.zeros((10,10)))
            return

        segmentation = self.model.predict(np.expand_dims(self.test_image, 0))
        label = np.argmax(np.squeeze(segmentation[0, :, :, :-1]), axis=2)
        self.fit_signal.emit(loss, val_loss, self.test_image[:, :, 0], label)


class ModelTrainer(QWidget, Ui_ModelTrainerUI):

    set_progress_signal = pyqtSignal(int, str)
    start_fitting_signal = pyqtSignal()
    end_fitting_signal = pyqtSignal()

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        self.setupUi(self)

        self.setWindowTitle('Dafne Model Trainer')
        self.advanced_widget.hide()
        self.fit_output_box.hide()
        self.adjustSize()
        self.pyplot_layout = QtWidgets.QVBoxLayout(self.fit_output_box)
        self.pyplot_layout.setObjectName("pyplot_layout")
        self.data_dir = None
        self.fig = plt.figure()
        self.fig.tight_layout()
        self.fig.set_tight_layout(True)
        self.canvas = FigureCanvas(self.fig)
        self.ax_left = self.fig.add_subplot(121)
        self.ax_left_twin = self.ax_left.twinx()
        self.ax_right = self.fig.add_subplot(122)
        self.ax_right.set_title('Current output')
        self.ax_right.axis('off')

        self.pyplot_layout.addWidget(self.canvas)
        self.bottom_widget = QWidget()
        bottom_layout = QtWidgets.QHBoxLayout(self.bottom_widget)

        bottom_layout.addWidget(QtWidgets.QLabel('Show validation slice:'))

        self.slice_select_slider = QtWidgets.QSlider(self.bottom_widget)
        self.slice_select_slider.setOrientation(QtCore.Qt.Horizontal)
        bottom_layout.addWidget(self.slice_select_slider)
        self.slice_select_slider.valueChanged.connect(self.val_slice_changed)
        self.slice_select_slider.setRange(0, 0)
        self.slice_select_slider.setEnabled(False)

        self.auto_stop_training_checkBox = QtWidgets.QCheckBox('Auto stop training', self.bottom_widget)
        self.auto_stop_training_checkBox.setChecked(True)
        self.auto_stop_training_checkBox.stateChanged.connect(self.auto_stop_training_changed)
        bottom_layout.addWidget(self.auto_stop_training_checkBox)

        self.pyplot_layout.addWidget(self.bottom_widget)

        self.advanced_button.clicked.connect(self.show_advanced)
        self.choose_Button.clicked.connect(self.choose_data)
        self.save_choose_Button.clicked.connect(self.choose_save_location)
        self.set_progress_signal.connect(self.set_progress)
        self.start_fitting_signal.connect(self.start_fitting_slot)
        self.end_fitting_signal.connect(self.stop_fitting_slot)
        self.fit_Button.clicked.connect(self.fit_clicked)
        self.preprocess_Button.clicked.connect(self.preprocess_clicked)
        self.force_preprocess_check.stateChanged.connect(self.decide_enable_fit)
        self.is_fitting = False
        self.fitting_ui_callback = None
        self.loss_list = []
        self.val_loss_list = []
        self.current_val_slice = 0
        self.val_image_list = []
        self.preprocessed_data_exist = False

    @pyqtSlot()
    def show_advanced(self):
        self.advanced_widget.show()
        self.advanced_button.hide()
        self.adjustSize()

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        if self.is_fitting:
            event.ignore()

    def decide_enable_fit(self):
        if self.preprocessed_data_exist and not self.force_preprocess_check.isChecked():
            self.fit_Button.setText('Fit')
        else:
            self.fit_Button.setText('Preprocess + fit')
        if self.location_Text.text() and self.model_location_Text.text():
            self.fit_Button.setEnabled(True)
            if not self.preprocessed_data_exist or self.force_preprocess_check.isChecked():
                self.preprocess_Button.setEnabled(True)
            else:
                self.preprocess_Button.setEnabled(False)
        else:
            self.fit_Button.setEnabled(False)
            self.preprocess_Button.setEnabled(False)
            self.set_progress(0, '')

    def decide_preprocess(self):
        if os.path.exists(os.path.join(self.data_dir, 'training_obj.pickle')):
            self.preprocessed_data_exist = True
            self.force_preprocess_check.setEnabled(True)
            self.force_preprocess_check.setChecked(False)
        else:
            self.preprocessed_data_exist = False
            self.force_preprocess_check.setEnabled(False)
            self.force_preprocess_check.setChecked(True)

    @pyqtSlot()
    def choose_data(self):
        def invalid():
            self.save_choose_Button.setEnabled(False)
            self.location_Text.setText("")
            self.decide_enable_fit()
            self.data_dir = None

        npz_file, _ = QFileDialog.getOpenFileName(self, "Choose data files", "", "Numpy files (*.npz);;All Files(*)")

        if not npz_file:
            invalid()
            return

        self.data_dir = os.path.dirname(npz_file)

        npz_files = [f for f in os.listdir(self.data_dir) if f.endswith('.npz')]
        if len(npz_files) > 0:
            self.save_choose_Button.setEnabled(True)
            self.location_Text.setText(self.data_dir)
            self.decide_preprocess()
            self.decide_enable_fit()
        else:
            invalid()
            # show a warning dialog
            QMessageBox.warning(self, "Warning", "No npz files found in the selected directory")

    @pyqtSlot()
    def choose_save_location(self):
        def invalid():
            self.save_choose_Button.setEnabled(False)
            self.model_location_Text.setText("")
            self.decide_enable_fit()

        fileName, _ = QFileDialog.getSaveFileName(self, "Save Model As", "",
                                                  "Model Files (*.model);;All Files (*)")
        if fileName:
            self.model_dir = os.path.dirname(fileName)
            self.model_name = os.path.basename(fileName)
            if self.model_name.endswith('.model'):
                self.model_name = self.model_name[:-6]
            self.model_location_Text.setText(fileName)
            self.decide_enable_fit()
        else:
            invalid()

    @pyqtSlot()
    @pyqtSlot(bool)
    @separate_thread_decorator
    def fit(self, preprocess_only=False):
        self.is_fitting = True
        self.start_fitting_signal.emit()
        self.set_progress_signal.emit(0, 'Loading data')
        data_list = load_data(self.data_dir)
        set_data_path(self.data_dir)

        self.set_progress_signal.emit(10, 'Getting model info')
        common_resolution, model_size, label_dict = get_model_info(data_list)

        levels = self.levels_spin.value()
        conv_layers = self.convlayers_spin.value()
        kernel_size = self.kernsize_spin.value()

        set_force_preprocess(self.force_preprocess_check.isChecked())

        self.set_progress_signal.emit(20, 'Creating model')
        source, model_uuid = create_model_source(self.model_name, common_resolution, model_size, label_dict, levels, conv_layers, kernel_size)

        # write the new model generator script
        with open(os.path.join(self.model_dir, f'generate_{self.model_name}_model.py'), 'w') as f:
            f.write(source)

        create_model_function, apply_model_function, incremental_learn_function = get_model_functions(source)
        model = create_model_function()

        n_datasets = len(data_list)
        if n_datasets < 10:
            validation_split = 0.2
        else:
            validation_split = 0.1

        n_validation = int(n_datasets * validation_split)

        if n_validation == 0:
            print("WARNING: No validation data will be used")

        validation_data_list = data_list[:n_validation]
        training_data_list = data_list[n_validation:]

        self.set_progress_signal.emit(30, 'Preparing data')

        print('preparing data')

        training_generator, steps, x_val_list, y_val_list = prepare_data(training_data_list, validation_data_list,
                                                                         common_resolution, model_size, label_dict)

        print(f'{x_val_list[0].shape=}')

        if preprocess_only:
            self.set_progress_signal.emit(100, 'Done')
            self.end_fitting_signal.emit()
            self.is_fitting = False
            QMessageBox.information(self, "Information", "Preprocess done")
            return

        self.set_progress_signal.emit(50, 'Training model')

        self.fitting_ui_callback = PredictionUICallback()
        self.val_image_list = x_val_list
        self.fitting_ui_callback.fit_signal.connect(self.update_plot)
        self.fitting_ui_callback.set_auto_stop_training(self.auto_stop_training_checkBox.isChecked())
        self.slice_select_slider.setMaximum(len(x_val_list) - 1)
        if len(x_val_list) > 0:
            self.fitting_ui_callback.set_test_image(x_val_list[0])
            self.slice_select_slider.setEnabled(True)
        else:
            self.slice_select_slider.setEnabled(False)

        trained_model, history = train_model(model, training_generator, steps, x_val_list, y_val_list, [self.fitting_ui_callback])
        if self.fitting_ui_callback.best_weights is not None:
            trained_model.set_weights(self.fitting_ui_callback.best_weights)

        self.fitting_ui_callback.deleteLater()
        self.fitting_ui_callback = None

        self.set_progress_signal.emit(90, 'Saving model')

        model_object = DynamicDLModel(model_uuid,
                                     create_model_function,
                                     apply_model_function,
                                     incremental_learn_function=incremental_learn_function,
                                     weights=trained_model.get_weights(),
                                     timestamp_id=int(time.time())
                                     )

        with open(os.path.join(self.model_dir, f'{self.model_name}.model'), 'wb') as f:
            model_object.dump(f)

        # save weights
        os.makedirs(os.path.join(self.model_dir, 'weights'), exist_ok=True)
        trained_model.save_weights(os.path.join(self.model_dir, 'weights', f'weights_{self.model_name}.hdf5'))
        self.set_progress_signal.emit(100, 'Done')
        self.end_fitting_signal.emit()
        self.is_fitting = False
        # open a message box to show the user the model was saved
        QMessageBox.information(None, "Information", f"Model saved successfully as {self.model_name}.model")

    @pyqtSlot(int)
    def auto_stop_training_changed(self, checked):
        if self.fitting_ui_callback is not None:
            self.fitting_ui_callback.set_auto_stop_training(checked > 0)

    @pyqtSlot()
    def val_slice_changed(self):
        if not self.fitting_ui_callback:
            return
        try:
            self.fitting_ui_callback.set_test_image(self.val_image_list[self.slice_select_slider.value()])
        except IndexError:
            print("Validation slice out of range")

    @pyqtSlot(float, float, np.ndarray, np.ndarray)
    def update_plot(self, loss, val_loss, image, label):
        # Define the number of segments (including the transparent color for zero)
        num_segments = 21

        # Create a list of colors with varying alpha values
        colors = [(0, 0, 0, 0)]  # Transparent color for zero
        for i in range(1, num_segments):
            # Generate distinguishable colors using HSV color space
            color = plt.cm.get_cmap('hsv', num_segments)(i)
            color = (color[0], color[1], color[2], 0.5)
            colors.append(color)

        # Create the colormap
        labels_colormap = ListedColormap(colors)

        self.loss_list.append(loss)
        self.val_loss_list.append(val_loss)

        y_axis_formatter = FuncFormatter(lambda y, _: '{:.1g}'.format(y))

        self.ax_left.clear()
        self.ax_left.plot(self.loss_list, color='#E66100', label='train')
        self.ax_left.set_ylabel('Training loss', color='#E66100')
        self.ax_left.yaxis.set_major_formatter(y_axis_formatter)
        self.ax_left_twin.clear()
        self.ax_left_twin.plot(self.val_loss_list, color='#5D3A9B', label='validation')
        self.ax_left_twin.set_ylabel('Validation loss', color='#5D3A9B')
        self.ax_left_twin.yaxis.set_label_position('right')
        self.ax_left_twin.yaxis.set_major_formatter(y_axis_formatter)
        self.ax_right.clear()
        self.ax_right.set_title('Current output')
        self.ax_right.imshow(image, cmap='gray')
        self.ax_right.imshow(label, cmap=labels_colormap)
        self.ax_right.axis('off')
        self.canvas.draw()

    @pyqtSlot(int, str)
    def set_progress(self, progress, message=''):
        self.progressBar.setValue(progress)
        self.progress_Label.setText(message)

    @pyqtSlot()
    def start_fitting_slot(self):
        self.loss_list = []
        self.val_loss_list = []
        self.fit_Button.setText('Stop fitting')
        self.preprocess_Button.setEnabled(False)
        self.fit_output_box.show()
        self.adjustSize()
        self.choose_Button.setEnabled(False)
        self.save_choose_Button.setEnabled(False)
        self.advanced_widget.setEnabled(False)
        self.advanced_button.setEnabled(False)
        if self.val_image_list:
            self.slice_select_slider.setEnabled(True)

    @pyqtSlot()
    def stop_fitting_slot(self):
        self.fit_Button.setText('Fit model')
        self.slice_select_slider.setEnabled(False)
        self.advanced_widget.setEnabled(True)
        self.advanced_button.setEnabled(True)
        self.decide_preprocess()
        self.decide_enable_fit()

    @pyqtSlot()
    def preprocess_clicked(self):
        self.fit(True)

    @pyqtSlot()
    def fit_clicked(self):
        if self.is_fitting:
            #stop the fitting
            if self.fitting_ui_callback is not None:
                self.fitting_ui_callback.stop()
                self.fit_Button.setText('Stopping...')
        else:
            self.fit()


def main():
    import sys
    app = QtWidgets.QApplication(sys.argv)
    app.setApplicationName('Dafne Model trainer')
    window = ModelTrainer()
    window.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ModelTrainer.ui'
#
# Created by: PyQt5 UI code generator 5.15.6
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_ModelTrainerUI(object):
    def setupUi(self, ModelTrainerUI):
        ModelTrainerUI.setObjectName("ModelTrainerUI")
        ModelTrainerUI.resize(815, 642)
        self.verticalLayout = QtWidgets.QVBoxLayout(ModelTrainerUI)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.label = QtWidgets.QLabel(ModelTrainerUI)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setObjectName("label")
        self.horizontalLayout.addWidget(self.label)
        self.location_Text = QtWidgets.QLineEdit(ModelTrainerUI)
        self.location_Text.setEnabled(False)
        self.location_Text.setReadOnly(True)
        self.location_Text.setObjectName("location_Text")
        self.horizontalLayout.addWidget(self.location_Text)
        self.choose_Button = QtWidgets.QPushButton(ModelTrainerUI)
        self.choose_Button.setObjectName("choose_Button")
        self.horizontalLayout.addWidget(self.choose_Button)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label_2 = QtWidgets.QLabel(ModelTrainerUI)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout_2.addWidget(self.label_2)
        self.model_location_Text = QtWidgets.QLineEdit(ModelTrainerUI)
        self.model_location_Text.setEnabled(False)
        self.model_location_Text.setReadOnly(True)
        self.model_location_Text.setObjectName("model_location_Text")
        self.horizontalLayout_2.addWidget(self.model_location_Text)
        self.save_choose_Button = QtWidgets.QPushButton(ModelTrainerUI)
        self.save_choose_Button.setEnabled(False)
        self.save_choose_Button.setObjectName("save_choose_Button")
        self.horizontalLayout_2.addWidget(self.save_choose_Button)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.force_preprocess_check = QtWidgets.QCheckBox(ModelTrainerUI)
        self.force_preprocess_check.setEnabled(False)
        self.force_preprocess_check.setChecked(True)
        self.force_preprocess_check.setObjectName("force_preprocess_check")
        self.horizontalLayout_3.addWidget(self.force_preprocess_check)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem)
        self.advanced_button = QtWidgets.QPushButton(ModelTrainerUI)
        self.advanced_button.setObjectName("advanced_button")
        self.horizontalLayout_3.addWidget(self.advanced_button)
        self.advanced_widget = QtWidgets.QWidget(ModelTrainerUI)
        self.advanced_widget.setEnabled(True)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.advanced_widget.sizePolicy().hasHeightForWidth())
        self.advanced_widget.setSizePolicy(sizePolicy)
        self.advanced_widget.setMinimumSize(QtCore.QSize(0, 0))
        self.advanced_widget.setObjectName("advanced_widget")
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout(self.advanced_widget)
        self.horizontalLayout_5.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_5.setSpacing(0)
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.label_3 = QtWidgets.QLabel(self.advanced_widget)
        self.label_3.setObjectName("label_3")
        self.horizontalLayout_5.addWidget(self.label_3)
        self.levels_spin = QtWidgets.QSpinBox(self.advanced_widget)
        self.levels_spin.setMinimum(1)
        self.levels_spin.setProperty("value", 5)
        self.levels_spin.setObjectName("levels_spin")
        self.horizontalLayout_5.addWidget(self.levels_spin)
        self.label_4 = QtWidgets.QLabel(self.advanced_widget)
        self.label_4.setObjectName("label_4")
        self.horizontalLayout_5.addWidget(self.label_4)
        self.convlayers_spin = QtWidgets.QSpinBox(self.advanced_widget)
        self.convlayers_spin.setProperty("value", 2)
        self.convlayers_spin.setObjectName("convlayers_spin")
        self.horizontalLayout_5.addWidget(self.convlayers_spin)
        self.label_5 = QtWidgets.QLabel(self.advanced_widget)
        self.label_5.setObjectName("label_5")
        self.horizontalLayout_5.addWidget(self.label_5)
        self.kernsize_spin = QtWidgets.QSpinBox(self.advanced_widget)
        self.kernsize_spin.setMinimum(1)
        self.kernsize_spin.setProperty("value", 2)
        self.kernsize_spin.setObjectName("kernsize_spin")
        self.horizontalLayout_5.addWidget(self.kernsize_spin)
        self.horizontalLayout_3.addWidget(self.advanced_widget)
        self.verticalLayout.addLayout(self.horizontalLayout_3)
        self.progressBar = QtWidgets.QProgressBar(ModelTrainerUI)
        self.progressBar.setEnabled(False)
        self.progressBar.setProperty("value", 0)
        self.progressBar.setObjectName("progressBar")
        self.verticalLayout.addWidget(self.progressBar)
        self.progress_Label = QtWidgets.QLabel(ModelTrainerUI)
        self.progress_Label.setText("")
        self.progress_Label.setObjectName("progress_Label")
        self.verticalLayout.addWidget(self.progress_Label)
        self.fit_output_box = QtWidgets.QGroupBox(ModelTrainerUI)
        self.fit_output_box.setMinimumSize(QtCore.QSize(800, 400))
        self.fit_output_box.setObjectName("fit_output_box")
        self.verticalLayout.addWidget(self.fit_output_box)
        spacerItem1 = QtWidgets.QSpacerItem(20, 45, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem1)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.preprocess_Button = QtWidgets.QPushButton(ModelTrainerUI)
        self.preprocess_Button.setEnabled(False)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.preprocess_Button.sizePolicy().hasHeightForWidth())
        self.preprocess_Button.setSizePolicy(sizePolicy)
        self.preprocess_Button.setObjectName("preprocess_Button")
        self.horizontalLayout_4.addWidget(self.preprocess_Button)
        self.fit_Button = QtWidgets.QPushButton(ModelTrainerUI)
        self.fit_Button.setEnabled(False)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(2)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.fit_Button.sizePolicy().hasHeightForWidth())
        self.fit_Button.setSizePolicy(sizePolicy)
        self.fit_Button.setObjectName("fit_Button")
        self.horizontalLayout_4.addWidget(self.fit_Button)
        self.verticalLayout.addLayout(self.horizontalLayout_4)

        self.retranslateUi(ModelTrainerUI)
        QtCore.QMetaObject.connectSlotsByName(ModelTrainerUI)

    def retranslateUi(self, ModelTrainerUI):
        _translate = QtCore.QCoreApplication.translate
        ModelTrainerUI.setWindowTitle(_translate("ModelTrainerUI", "Form"))
        self.label.setText(_translate("ModelTrainerUI", "Data location"))
        self.choose_Button.setText(_translate("ModelTrainerUI", "Choose..."))
        self.label_2.setText(_translate("ModelTrainerUI", "Output model:"))
        self.save_choose_Button.setText(_translate("ModelTrainerUI", "Choose..."))
        self.force_preprocess_check.setText(_translate("ModelTrainerUI", "Force preprocess"))
        self.advanced_button.setText(_translate("ModelTrainerUI", "Advanced settings"))
        self.label_3.setText(_translate("ModelTrainerUI", "Levels: "))
        self.label_4.setText(_translate("ModelTrainerUI", " Conv. layers: "))
        self.label_5.setText(_translate("ModelTrainerUI", " Kernel size: "))
        self.fit_output_box.setTitle(_translate("ModelTrainerUI", "Fitting output"))
        self.preprocess_Button.setText(_translate("ModelTrainerUI", "Preprocess only"))
        self.fit_Button.setText(_translate("ModelTrainerUI", "Preprocess + Fit model"))




from PyQt5.QtCore import QRunnable, pyqtSlot, QThreadPool
from functools import wraps
import traceback

threadpool = QThreadPool()

class Runner(QRunnable):

    def __init__(self, func, *args, **kwargs):
        QRunnable.__init__(self)
        self.func = func
        self.args = args
        self.kwargs = kwargs

    @pyqtSlot()
    def run(self):
        try:
            setattr(self.args[0], 'separate_thread_running', True)
        except:
            pass
        self.func(*self.args, **self.kwargs)
        try:
            setattr(self.args[0], 'separate_thread_running', False)
        except:
            pass

def separate_thread_decorator(func):
    @wraps(func)
    def run_wrapper(*args, **kwargs):
        runner = Runner(func, *args, **kwargs)
        threadpool.start(runner)
    return run_wrapper


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import os
import pickle
import sys
import time
import uuid

from dafne_dl.DynamicDLModel import source_to_fn, DynamicDLModel

assert sys.version_info.major == 3, "This software is only compatible with Python 3.x"

if sys.version_info.minor < 9:
    from ..utils.source_tools import extract_function_source_basic as extract_function_source
else:
    from ..utils.source_tools import extract_function_source

from .. import resources
if sys.version_info.minor < 10:
    import importlib_resources as pkg_resources
else:
    import importlib.resources as pkg_resources

import matplotlib.pyplot as plt
import numpy as np
from dafne_dl.common.DataGenerators import DataGeneratorMem
from dafne_dl.common.preprocess_train import common_input_process_single, input_creation_mem, weighted_loss
from dafne_dl.labels.utils import invert_dict
from tensorflow.keras import optimizers
from tensorflow.keras.callbacks import Callback

BATCH_SIZE = 5
MAX_EPOCHS = 400
PATIENCE = 5
ENABLE_GUI = False
STORE_PREPROCESS = True
FORCE_PREPROCESS = False
PREPROCESS_ONLY = False
MIN_EPOCHS = 10
MAX_INPUT_SIZE = 256
DEFAULT_BIASCORRECTION_LEVELS = 4
DEFAULT_BIASCORRECTION_NORMALIZE = -1

DATA_PATH = None


def set_data_path(path):
    global DATA_PATH
    DATA_PATH = path


def get_data_path():
    return DATA_PATH


def set_force_preprocess(force):
    global FORCE_PREPROCESS
    FORCE_PREPROCESS = force


def get_force_preprocess():
    return FORCE_PREPROCESS


# Get the path of the script's parent directory
parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir))
# Add the parent directory to sys.path to import common
sys.path.append(parent_dir)


def load_data(data_path):
    """
    Load all the npz files in the folder
    :param data_path: string containing the path to the folder
    :return: list of numpy file objects
    """
    # Load all the npz files in the folder
    data_files = [os.path.join(data_path, f) for f in os.listdir(data_path) if f.endswith('.npz')]
    data_list = []
    for file in data_files:
        print(f'Loading {file}')
        npz_file = np.load(file)
        data_list.append(npz_file)
    return data_list


def sanitize_label(label):
    if label.endswith('_L'):
        label = label[:-2] + '_X'
    elif label.endswith('_R'):
        label = label[:-2] + '_Y'
    return label


def get_model_info(data_list):
    """
    Get the common resolution, size and labels from the data
    :param data_list: the list of numpy file objects
    :return: common_resolution, model_size, label_dict
    """
    # Find the most common resolution among all the data
    resolutions = [data['resolution'][:2] for data in data_list]
    common_resolution = np.median(resolutions, axis=0)[:2]

    # find the maximum size normalized to the common resolution
    max_size = [np.ceil(data['data'].shape[:2]*data['resolution'][:2] / common_resolution).max() for data in data_list]
    max_size = int(max(max_size))
    if max_size > MAX_INPUT_SIZE:
        zoom_factor = MAX_INPUT_SIZE / max_size
        common_resolution = common_resolution / zoom_factor
        max_size = MAX_INPUT_SIZE
    model_size = (max_size, max_size)

    # Create a set with all the different labels present in the files
    labels = set()
    for data in data_list:
        for key in data.keys():
            if key.startswith('mask_'):
                label = key[5:]
                labels.add(label)

    labels = sorted(labels)

    n_labels = len(labels)
    # find datasets with missing labels
    missing_data_list = []
    for i, data in enumerate(data_list):
        n_existing_labels = 0
        for label in labels:
            if f'mask_{label}' not in data.keys() or data[f'mask_{label}'].sum() < 5:
                print(f'Warning: dataset {i} is missing label {label}')
            else:
                n_existing_labels += 1
        if n_existing_labels <= n_labels/2:
            missing_data_list.append(i)

    # remove datasets with missing labels
    for i in missing_data_list[::-1]:
        data_list.pop(i)

    # Crete a dictionary with indices for each label
    label_dict = {i+1: sanitize_label(label) for i, label in enumerate(labels)}
    return common_resolution, model_size, label_dict


def escape_string(string):
    """
    Escape a string to be used in a python source code
    :param string: the string to escape
    :return: the escaped string
    """
    return string.replace('"', '\\"')


def get_model_functions(source):
    """
    Extract the create_model function from the source code
    :param source: source in string form
    :return: functions returning a keras model, the apply function, and the incremental learn function
    """

    # source_to_fn automatically appends the source code as an attribute of the function
    model_create_fn = source_to_fn(extract_function_source(source, 'make_unet'))
    model_apply_fn = source_to_fn(extract_function_source(source, 'model_apply'))
    model_incremental_mem_fn = source_to_fn(extract_function_source(source, 'model_incremental_mem'))
    return model_create_fn, model_apply_fn, model_incremental_mem_fn


def convert_3d_mask_to_slices(mask_dictionary):
    """
    Convert a dictionary of 3d masks to list of dictionaries of 2d masks
    :param mask_dictionary: dictionary in the form {label: mask_3d}
    :return: a list of dictionaries in the form {label: mask_2d}
    """
    mask_list = []
    for i in range(mask_dictionary[list(mask_dictionary.keys())[0]].shape[2]):
        mask_list.append({sanitize_label(label): (mask[:, :, i] > 0).astype(np.uint8) for label, mask in mask_dictionary.items()})
    return mask_list


def normalize_training_data(data_list, common_resolution, model_size, label_dict):
    """
    Normalize the training data
    :param data_list: the list of numpy file objects
    :param common_resolution: the common resolution of the data
    :param model_size: the common size of the data
    :param label_dict: the dictionary with the labels
    :return: list of normalized data, list of normalized masks
    """

    inverse_label_dict = invert_dict(label_dict)

    all_slice_list = []
    all_masks_list = []

    for i,data in enumerate(data_list):
        print('Normalizing', i+1, '/', len(data_list))
        img_3d = data['data']
        image_list = [img_3d[:, :, i].astype(np.float32) for i in range(img_3d.shape[2])]
        training_data_dict = {'image_list': image_list, 'resolution': data['resolution'][:2]}
        mask_dictionary = {key[5:]: data[key] for key in data.keys() if key.startswith('mask_')}
        mask_list = convert_3d_mask_to_slices(mask_dictionary)

        processed_image_list, processed_mask_list = common_input_process_single(inverse_label_dict,
                                                                                common_resolution,
                                                                                model_size,
                                                                                model_size,
                                                                                training_data_dict,
                                                                                mask_list,
                                                                                False,
                                                                                DEFAULT_BIASCORRECTION_LEVELS,
                                                                                DEFAULT_BIASCORRECTION_NORMALIZE)
        all_slice_list.extend(processed_image_list)
        all_masks_list.extend(processed_mask_list)

    return all_slice_list, all_masks_list


def generate_training_and_weights(data_list, mask_list, band=49):
    return input_creation_mem(data_list, mask_list, band=band)


def make_validation_list(data_list, common_resolution, model_size, label_dict):
    """
    Create a validation list
    :param data_list: the list of datasets
    :param common_resolution: resolution of the model
    :param model_size: size of the model
    :param label_dict: dictionary of the labels
    :return:
    """
    if STORE_PREPROCESS and DATA_PATH and os.path.exists(os.path.join(DATA_PATH,'validation_obj.pickle')) and not FORCE_PREPROCESS:
        with open(os.path.join(DATA_PATH,'validation_obj.pickle'), 'rb') as f:
            training_objects = pickle.load(f)
    else:
        normalized_data_list, normalized_mask_list = normalize_training_data(data_list,
                                                                             common_resolution,
                                                                             model_size, label_dict)
        if not normalized_data_list:
            print('Warning! No valid validation data found!')
            return [], []
        training_objects = generate_training_and_weights(normalized_data_list, normalized_mask_list)
        if STORE_PREPROCESS and DATA_PATH:
            with open(os.path.join(DATA_PATH, 'validation_obj.pickle'), 'wb') as f:
                pickle.dump(training_objects, f)
    x_list = [np.stack([training_object[:,:,0], training_object[:,:,-1]], axis=-1) for training_object in training_objects]
    y_list = [training_object[:,:,1:-1] for training_object in training_objects]
    #plt.imshow(x_list[0][:,:,0])
    #plt.figure()
    #plt.imshow(y_list[0][:,:,2])
    #plt.figure()
    #plt.imshow(x_list[0][:,:,1])
    #plt.show()
    return x_list, y_list


def make_data_generator(data_list, common_resolution, model_size, label_dict):
    """
    Create a data generator
    :param data_list: the list of datasets
    :param common_resolution: resolution of the model
    :param model_size: size of the model
    :param label_dict: dictionary of the labels
    :return:
    """
    if STORE_PREPROCESS and DATA_PATH and os.path.exists(os.path.join(DATA_PATH,'training_obj.pickle')) and not FORCE_PREPROCESS:
        with open(os.path.join(DATA_PATH,'training_obj.pickle'), 'rb') as f:
            training_objects = pickle.load(f)
    else:
        print("Normalizing data...")
        normalized_data_list, normalized_mask_list = normalize_training_data(data_list,
                                                                             common_resolution,
                                                                             model_size, label_dict)
        print("Generating training objects...")
        training_objects = generate_training_and_weights(normalized_data_list, normalized_mask_list)
        if STORE_PREPROCESS and DATA_PATH: 
            with open(os.path.join(DATA_PATH, 'training_obj.pickle'), 'wb') as f:
                pickle.dump(training_objects, f)
    steps = int(len(training_objects) / BATCH_SIZE)
    data_generator = DataGeneratorMem(training_objects, list_X=list(range(steps * BATCH_SIZE)),
                                               batch_size=BATCH_SIZE, dim=model_size)

    return data_generator, steps


def prepare_data(training_data_list, validation_data_list, common_resolution, model_size, label_dict):
    training_generator, steps = make_data_generator(training_data_list, common_resolution, model_size, label_dict)

    if len(validation_data_list) > 0:
        x_val_list, y_val_list = make_validation_list(validation_data_list, common_resolution, model_size, label_dict)
    else:
        x_val_list = []
        y_val_list = []

    return training_generator, steps, x_val_list, y_val_list


def train_model(model, training_generator, steps, x_val_list, y_val_list, custom_callbacks=None):
    """
    Train the model
    :param model: Keras model
    :param data_list: list of data
    :param common_resolution: resolution of the model
    :param model_size: size of the model
    :param label_dict: dictionary with labels
    :return: the trained model
    """

    n_validation = len(x_val_list)

    # now train the model on the data
    adamlr = optimizers.Adam(learning_rate=0.009765, beta_1=0.9, beta_2=0.999, epsilon=1e-08, amsgrad=True)
    model.compile(loss=weighted_loss, optimizer=adamlr)

    # do the training
    if n_validation > 0:
        if ENABLE_GUI:
            plt.ion()
        class PredictionCallback(Callback):
            def __init__(self):
                super().__init__()
                self.min_val_loss = np.inf
                self.n_val_loss_increases = 0
                self.best_weights = None
            def on_epoch_end(self, epoch, logs=None):
                if epoch < MIN_EPOCHS: return
                if logs['val_loss'] < self.min_val_loss:
                    self.min_val_loss = logs['val_loss']
                    self.n_val_loss_increases = 0
                    self.best_weights = self.model.get_weights()
                elif logs['val_loss'] > self.min_val_loss:
                    self.n_val_loss_increases += 1

                if self.n_val_loss_increases >= PATIENCE:
                    self.model.stop_training = True

                if ENABLE_GUI:
                    segmentation = self.model.predict(np.expand_dims(x_val_list[0],0))
                    #plt.imshow(x_val_list[0][:,:,0])
                    #plt.figure()
                    label = np.argmax(np.squeeze(segmentation[0, :, :, :-1]), axis=2)
                    plt.imshow(label)
                    plt.show(block=False)
                    plt.pause(0.001)

        prediction_callback = PredictionCallback()

        if custom_callbacks is None:
            custom_callbacks = [prediction_callback]

        history = model.fit(training_generator, epochs=MAX_EPOCHS,
                  steps_per_epoch=steps,
                  validation_data=(np.stack(x_val_list,0), np.stack(y_val_list,0)),
                  callbacks=custom_callbacks,
                  verbose=1)

        if prediction_callback.best_weights is not None:
            model.set_weights(prediction_callback.best_weights)

    else:
        if custom_callbacks is None:
            custom_callbacks = []
        history = model.fit(training_generator, epochs=MAX_EPOCHS, steps_per_epoch=steps, verbose=1, callbacks=custom_callbacks)

    return model, history


def create_model_source(model_name, common_resolution, model_size, label_dict, levels=5, conv_layers=2, kernel_size=2):
    """
    Create the source code of the model
    :param common_resolution: resolution of the model
    :param model_size: size of the model
    :param label_dict: dictionary of the labels
    :param levels: number of levels
    :return: the source code
    """
    # load the model template
    if getattr(sys, '_MEIPASS', None): # PyInstaller support. If _MEIPASS is set, we are in a Pyinstaller environment
        with open(os.path.join(sys._MEIPASS, 'resources', 'generate_model.py.tmpl'), 'r') as f:
            source = f.read()
    else:
        with pkg_resources.files(resources).joinpath('generate_model.py.tmpl').open() as f:
            source = f.read()

    model_uuid = str(uuid.uuid4())

    # replace the variables
    source = source.replace('%%MODEL_NAME%%', f'"{escape_string(model_name)}"')
    source = source.replace('%%MODEL_RESOLUTION%%', f'[{common_resolution[0]:.2f}, {common_resolution[1]:.2f}]')
    source = source.replace('%%MODEL_SIZE%%', f'[{model_size[0]:d}, {model_size[1]:d}]')
    source = source.replace('%%MODEL_UID%%', f'"{model_uuid}"')
    source = source.replace('%%LABELS_DICT%%', str(label_dict))
    source = source.replace('%%N_LEVELS%%', str(int(levels)))
    source = source.replace('%%N_CONV_LAYERS%%', str(int(conv_layers)))
    source = source.replace('%%INITIAL_KERNEL_SIZE%%', str(int(kernel_size)))
    source = source.replace('%%BIASCORRECTION_LEVELS%%', str(int(DEFAULT_BIASCORRECTION_LEVELS)))
    source = source.replace('%%BIASCORRECTION_NORMALIZE%%', str(int(DEFAULT_BIASCORRECTION_NORMALIZE)))

    return source, uuid


def create_model(model_name, data_path, levels=5, conv_layers=2, kernel_size=2, test_create_model=False):
    global DATA_PATH

    if test_create_model:
        print("Testing model creation")
        common_resolution = [0.5, 0.5]
        model_size = [321, 321]
        label_dict = {1: 'LK', 2:'RK'}
    else:
        data_list = load_data(data_path)
        DATA_PATH = data_path

        common_resolution, model_size, label_dict = get_model_info(data_list)

    source, model_id = create_model_source(model_name, common_resolution, model_size, label_dict, levels, conv_layers, kernel_size)

    # write the new model generator script
    with open(f'generate_{model_name}_model.py', 'w') as f:
        f.write(source)

    create_model_function, apply_model_function, incremental_learn_function = get_model_functions(source)
    model = create_model_function()

    if test_create_model:
        return None, None

    n_datasets = len(data_list)
    if n_datasets < 10:
        validation_split = 0.2
    else:
        validation_split = 0.1
    n_validation = int(n_datasets * validation_split)

    if n_validation == 0:
        print("WARNING: No validation data will be used")

    validation_data_list = data_list[:n_validation]
    training_data_list = data_list[n_validation:]

    training_generator, steps, x_val_list, y_val_list = prepare_data(training_data_list, validation_data_list,
                                                                     common_resolution, model_size, label_dict)

    if PREPROCESS_ONLY:
        return None, None

    trained_model, history = train_model(model, training_generator, steps, x_val_list, y_val_list)

    model_object = DynamicDLModel(model_id,
                                  create_model_function,
                                  apply_model_function,
                                  incremental_learn_function=incremental_learn_function,
                                  weights=trained_model.get_weights(),
                                  timestamp_id=int(time.time())
                                  )

    return model_object, history


def save_weights(model, model_name):
    """
    Save the weights of the model
    :param model: the model
    :param model_name: the name of the model
    :return:
    """
    os.makedirs('weights', exist_ok=True)
    model_path = os.path.join('weights', f'weights_{model_name}.hdf5')
    model.save_weights(model_path)
    print(f'Saved weights to {model_path}')


def main():
    global ENABLE_GUI, FORCE_PREPROCESS, PREPROCESS_ONLY, MIN_EPOCHS

    parser = argparse.ArgumentParser()
    parser.add_argument("model_name", help="Name of the model to create")
    parser.add_argument("data_path", help="Path to data folder (containing the '*.npz' files)")
    parser.add_argument("--no-gui", "-n", help="Disable the GUI", action="store_true")
    parser.add_argument("--levels", "-l", help="Number of levels of the model", type=int, default=5)
    parser.add_argument("--conv-layers", "-c", help="Number of convolutional layers", type=int, default=2)
    parser.add_argument("--kernel-size", "-k", help="Initial size of the kernel", type=int, default=2)
    parser.add_argument("--preprocess-only", "-p", help="Only preprocess the data", action="store_true")
    parser.add_argument("--force-preprocess", "-f", help="Force preprocessing of the data", action="store_true")
    parser.add_argument("--min-epochs", "-m", help="Minimum number of epochs", type=int, default=10)
    parser.add_argument("--test-model", "-t", help="Test the model creation", action="store_true")
    args = parser.parse_args()

    if args.no_gui:
        ENABLE_GUI = False

    if args.force_preprocess:
        FORCE_PREPROCESS = True

    if args.preprocess_only:
        PREPROCESS_ONLY = True

    MIN_EPOCHS = args.min_epochs

    dl_model, history = create_model(args.model_name, args.data_path, args.levels, args.conv_layers, args.kernel_size, args.test_model)

    if PREPROCESS_ONLY or args.test_model:
        return

    save_weights(dl_model.model, args.model_name)

    with open(f'{args.model_name}.model', 'wb') as f:
        dl_model.dump(f)

    if ENABLE_GUI:
        plt.plot(history.history['loss'])
        try:
            plt.plot(history.history['val_loss'])
            plt.legend(['train', 'validation'], loc='upper left')
        except KeyError:
            # no validation
            pass
        plt.show()


if __name__ == '__main__':
    main()



# setup.py
from setuptools import setup
import subprocess
import os

BASE_PACKAGE_NAME = 'dafne-models'

REQUIREMENTS = {
    '*': [
        'matplotlib',
        'ormir-pyvoxel',
        'dafne-dicomUtils >= 0.1.7',
        'dill'
    ],
    'main': [
        'dafne_dl >= 1.4a2'
    ],
    'torch_models': [
        'dafne_dl-torch_models >= 2.0a2'
    ]
}


def get_current_branch():
    """Get the current git branch name."""
    # First try environment variable (for CI/CD environments)
    branch = os.getenv('GIT_BRANCH')
    if branch:
        return branch.split('/')[-1]  # Handle cases like 'origin/main'

    # Then try git command
    try:
        branch = subprocess.check_output(
            ['git', 'rev-parse', '--abbrev-ref', 'HEAD'],
            stderr=subprocess.DEVNULL  # Suppress stderr
        ).decode('utf-8').strip()
        return branch
    except (subprocess.CalledProcessError, FileNotFoundError):
        return 'main'  # Default fallback


def get_package_name():
    """Determine package name based on git branch."""
    branch = get_current_branch()

    if branch == 'main':
        return BASE_PACKAGE_NAME
    else:
        return f'{BASE_PACKAGE_NAME}-{branch}'

def get_requirements():
    """Get the requirements for the current branch."""
    branch = get_current_branch()
    requirements_common = REQUIREMENTS.get('*', [])
    reqs = requirements_common + REQUIREMENTS.get(branch, REQUIREMENTS['main'])
    print(reqs)
    return reqs

setup(
    name=get_package_name(),
    install_requires=get_requirements()
    # Other setup parameters will be read from setup.cfg
)


#!/bin/env python3
# -*- coding: utf-8 -*-
# tested models for leg: 1610001000 (initial), 1669385545 (final)

import argparse
import numpy as np
import os
import voxel as vx
from dicomUtils import medical_volume_from_path, realign_medical_volume
from dafne_dl.model_loaders import generic_load_model
import flexidep
import ast
import pprint
import dill

APP_STRING = 'network.dafne_models.run_model'

def parse_complex_options(options_string):
    def parse_value(value):
        value = value.strip()
        if value.lower() == 'true':
            return True
        elif value.lower() == 'false':
            return False
        try:
            return ast.literal_eval(value)
        except (ValueError, SyntaxError):
            return value

    options = {}
    current_key = None
    current_value = ''
    stack = []
    in_quotes = False
    quote_char = None

    for char in options_string:
        if char in ('"', "'") and (not in_quotes or char == quote_char):
            in_quotes = not in_quotes
            quote_char = char if in_quotes else None
            current_value += char
        elif char == ',' and not stack and not in_quotes:
            if current_key:
                options[current_key] = parse_value(current_value)
                current_key = None
                current_value = ''
        elif char == '=' and not stack and not in_quotes:
            current_key = current_value.strip()
            current_value = ''
        elif char == '[' and not in_quotes:
            stack.append('[')
            current_value += char
        elif char == ']' and not in_quotes:
            if stack and stack[-1] == '[':
                stack.pop()
            current_value += char
        else:
            current_value += char

    if current_key:
        options[current_key] = parse_value(current_value)

    return options

def main():
    parser = argparse.ArgumentParser()

    parser.add_argument("image_path", help="Path to image/dataset to segment")
    parser.add_argument("other_contrasts", nargs='*', help="Other contrasts to segment")
    parser.add_argument("--model", "-m", type=str, required=True, help="Path to the model file to use")
    parser.add_argument("--classification", nargs=1, metavar='class', type=str, required=False, help="Classification label")
    parser.add_argument("--output", "-o", type=str, required=False, help="Path to save the output", default='.')
    parser.add_argument("--options", "-p", type=str, required=False, help="Options to be passed to the model", default='')
    args = parser.parse_args()

    options=parse_complex_options(args.options)

    try:
        model_classification = args.classification[0].split(',')[0]
    except TypeError:
        model_classification = ''

    print("Loading model")

    with open(args.model, 'rb') as f:
        input_dict = dill.load(f)

    metadata = input_dict.get('metadata', {})
    print("Model metadata:")
    pprint.pp(metadata)

    # check and install model dependencies
    dependencies = metadata.get('dependencies', {})

    dependency_manager = flexidep.DependencyManager(
        config_file=None,
        config_string=None,
        unique_id=APP_STRING,
        interactive_initialization=False,
        use_gui=False,
        install_local=False,
        package_manager=flexidep.PackageManagers.pip,
        extra_command_line='',
    )

    for package, alternative_str in dependencies.items():
        print("Processing package", package)
        dependency_manager.process_single_package(package, alternative_str)

    # build the model object. Now that the dependencies are installed, the model can be loaded
    model = generic_load_model(input_dict)

    # check if the model has a specific orientation
    model_orientation = metadata.get('orientation', None)

    if isinstance(model_orientation, str):
        # the orientation is a string (Axial/Transversal, Sagittal, Coronal)
        model_orientation = model_orientation.lower()
        if model_orientation.startswith('a') or model_orientation.startswith('t'):
            model_orientation = ('LR', 'AP', 'SI')
        elif model_orientation.startswith('s'):
            model_orientation = ('AP', 'IS', 'LR')
        elif model_orientation.startswith('c'):
            model_orientation = ('LR', 'SI', 'AP')
        else:
            print("Unknown orientation")
            model_orientation = None

    image = medical_volume_from_path(args.image_path, reorient_data=False)

    original_orientation = image.orientation

    if model_orientation is not None and model_orientation != original_orientation:
        image.reformat(model_orientation, inplace=True)

    resolution = image.pixel_spacing
    inputs = [image.volume.astype(np.float32)]

    print("Image loaded")

    for contrast in args.other_contrasts:
        contrast_image = medical_volume_from_path(contrast, reorient_data=False)
        contrast_image = realign_medical_volume(contrast_image, image)
        inputs.append(contrast_image.volume.astype(np.float32))

    print("Contrasts loaded")

    dimensionality = model.data_dimensionality
    output_masks = {}
    if dimensionality == 2: # this is a 2D model
        print("2D model")
        n_slices = image.shape[2]
        for i in range(n_slices):
            input_dict = {'image': inputs[0][:, :, i], 'resolution': resolution[:2], 'options': options, 'split_laterality': False, 'classification': model_classification}
            for idx, contrast in enumerate(inputs[1:]):
                input_dict[f'image{idx+1}'] = contrast[:, :, i]
            output = model.apply(input_dict)
            for key, mask in output.items():
                if key not in output_masks:
                    output_masks[key] = np.zeros((image.shape[0], image.shape[1], n_slices), dtype=np.uint8)
                output_masks[key][:, :, i] = mask
    else: # this is a 3D model
        print("3D model")
        input_dict = {'image': inputs[0], 'resolution': resolution, 'options': options, 'split_laterality': False, 'classification': model_classification}
        for idx, contrast in enumerate(inputs[1:]):
            print("Adding contrast", f'image{idx+2}')
            input_dict[f'image{idx+2}'] = contrast
        print("Applying model")
        output = model.apply(input_dict)
        for key, mask in output.items():
            output_masks[key] = mask

    writer = vx.NiftiWriter()

    for key, mask in output_masks.items():
        output_data = vx.MedicalVolume(mask, image.affine)
        if model_orientation is not None and model_orientation != original_orientation:
            output_data.reformat(original_orientation, inplace=True)
        writer.save(output_data, os.path.join(args.output, f'{key}.nii.gz'))


if __name__ == '__main__':
    main()